diff --git a/res/res_pjsip_session.c b/res/res_pjsip_session.c
index d66a819..5beea88 100644
--- a/res/res_pjsip_session.c
+++ b/res/res_pjsip_session.c
@@ -2353,6 +2353,21 @@ static void handle_incoming_request(struct ast_sip_session *session, pjsip_rx_da
 	}
 }
 
+static void session_end(struct ast_sip_session *session)
+{
+	struct ast_sip_session_supplement *iter;
+
+	/* Stop the scheduled termination */
+	sip_session_defer_termination_stop_timer(session);
+
+	/* Session is dead.  Notify the supplements. */
+	AST_LIST_TRAVERSE(&session->supplements, iter, next) {
+		if (iter->session_end) {
+			iter->session_end(session);
+		}
+	}
+}
+
 static void handle_incoming_response(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_event_id_e type,
 		enum ast_sip_session_response_priority response_priority)
 {
@@ -2369,6 +2384,9 @@ static void handle_incoming_response(struct ast_sip_session *session, pjsip_rx_d
 		if (supplement->incoming_response && does_method_match(&rdata->msg_info.cseq->method.name, supplement->method)) {
 			supplement->incoming_response(session, rdata);
 		}
+		if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
+			session_end(session);
+		}
 	}
 }
 
@@ -2427,21 +2445,6 @@ static void handle_outgoing(struct ast_sip_session *session, pjsip_tx_data *tdat
 	}
 }
 
-static void session_end(struct ast_sip_session *session)
-{
-	struct ast_sip_session_supplement *iter;
-
-	/* Stop the scheduled termination */
-	sip_session_defer_termination_stop_timer(session);
-
-	/* Session is dead.  Notify the supplements. */
-	AST_LIST_TRAVERSE(&session->supplements, iter, next) {
-		if (iter->session_end) {
-			iter->session_end(session);
-		}
-	}
-}
-
 /*!
  * \internal
  * \brief Complete ending session activities.
@@ -2468,6 +2471,7 @@ static void session_inv_on_state_changed(pjsip_inv_session *inv, pjsip_event *e)
 {
 	struct ast_sip_session *session = inv->mod_data[session_module.id];
 	pjsip_event_id_e type;
+	int response_received = 0;
 
 	if (e) {
 		print_debug_details(inv, NULL, e);
@@ -2487,6 +2491,9 @@ static void session_inv_on_state_changed(pjsip_inv_session *inv, pjsip_event *e)
 	case PJSIP_EVENT_RX_MSG:
 		handle_incoming(session, e->body.rx_msg.rdata, type,
 				AST_SIP_SESSION_BEFORE_MEDIA);
+		if (e->body.rx_msg.rdata->msg_info.msg->type == PJSIP_RESPONSE_MSG) {
+			response_received = 1;
+		}
 		break;
 	case PJSIP_EVENT_TSX_STATE:
 		ast_debug(3, "Source of transaction state change is %s\n", pjsip_event_str(e->body.tsx_state.type));
@@ -2498,6 +2505,9 @@ static void session_inv_on_state_changed(pjsip_inv_session *inv, pjsip_event *e)
 		case PJSIP_EVENT_RX_MSG:
 			handle_incoming(session, e->body.tsx_state.src.rdata, type,
 					AST_SIP_SESSION_BEFORE_MEDIA);
+			if (e->body.tsx_state.src.rdata->msg_info.msg->type == PJSIP_RESPONSE_MSG) {
+				response_received = 1;
+			}
 			break;
 		case PJSIP_EVENT_TRANSPORT_ERROR:
 		case PJSIP_EVENT_TIMER:
@@ -2516,7 +2526,7 @@ static void session_inv_on_state_changed(pjsip_inv_session *inv, pjsip_event *e)
 		break;
 	}
 
-	if (inv->state == PJSIP_INV_STATE_DISCONNECTED) {
+	if (inv->state == PJSIP_INV_STATE_DISCONNECTED && !response_received) {
 		session_end(session);
 	}
 }
